!========================================================================
! Copyright (C) GemTalk Systems 1986-2015.  All Rights Reserved.
!
! $Id: rcpipe.gs,v 1.7 2008-01-09 19:02:36 stever Exp $
!
! Superclass Hierarchy: 
!   RcPipe, Collection, Object.
!
!========================================================================

run
Object subclass: 'RcPipe'
  instVarNames: #( 'values' 'consumerState' 'producerState')
  classVars: #()
  classInstVars: #()
  poolDictionaries: #()
  inDictionary: Globals
  constraints: #()
  instancesInvariant: false
  isModifiable: false
  "reservedOop: 1657" .
  true
%
removeallmethods RcPipe
removeallclassmethods RcPipe

! ------------------- Class methods for RcPipe
category: 'For Documentation Installation only'
classmethod: RcPipe
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'An RcPipe is an implementation of a FIFO queue that provides significantly
 reduced concurrency conflicts when used in an environment with a single
 producer (a user who adds elements to the queue) and a single consumer
 (a user who may remove items from the queue).  In this environment, the
 producer and the consumer are guaranteed not to conflict with each other.

 Since it is a composite object with logical dependencies between its
 sub-components, it should not be used in a system where read-write conflicts
 are disabled.  If read-write conflicts are disabled, the dependencies may not
 be honored and a logically inconsistent state can result.' .
doc documentClassWith: txt.

txt := (GsDocText new) details:
'For GemStone internal use.' .
doc documentInstVar: #values with: txt.

txt := (GsDocText new) details:
'For GemStone internal use.' .
doc documentInstVar: #consumerState with: txt.

txt := (GsDocText new) details:
'For GemStone internal use.' .
doc documentInstVar: #producerState with: txt.

self description: doc.
%

category: 'Instance Creation'
classmethod: RcPipe
new

"Returns a new RcPipe."

^super new initialize: 10.
%

category: 'Instance Creation'
classmethod: RcPipe
new: anInitialSize

"Returns a new RcPipe. The argument anInitialSize provides a hint of the number
 of elements that the instance should be designed to hold."

^super new initialize: anInitialSize
%

! ------------------- Instance methods for RcPipe
category: 'Accessing'
method: RcPipe
size

"Returns the number of valid entries in the RcPipe."

|v|

v := self _firstValidOffset.
^(values size - v) + 1
%

category: 'Adding'
method: RcPipe
add: aValue

"Adds aValue to the RcPipe, returns the receiver."

self _setupAdd.
values add: aValue.
^self
%

category: 'Adding'
method: RcPipe
addAll: aCollection

"Adds all of the elements of aCollection, returns the receiver."

self _setupAdd.
values addAll: aCollection.
^self
%

category: 'Converting'
method: RcPipe
asArray

"Returns an Array with the contents of the receiver."

^Array withAll: self
%

category: 'Comparing'
method: RcPipe
= anRcPipe

"Returns true if all of the following conditions are true:

 1.  The receiver and anRcPipe are of the same class.
 2.  The two pipes are the same size.
 3.  The corresponding elements of the receiver and anRcPipe
     are equal."

System _addRootObjectToRcReadSet: self.
System _addEntireObjectToRcReadSet: values.

self == anRcPipe ifTrue: [ ^true ].

anRcPipe class == self class
    ifFalse: [ ^ false ].

self size = anRcPipe size ifFalse: [ ^false ].

^self asArray = anRcPipe asArray
%

category: 'Enumerating'
method: RcPipe
do: aBlock

"Evaluates aBlock with each of the current elements of the RcPipe as
 the argument. The argument aBlock must be a one-argument block.  This
 method traverses the queue elements in order."

|result|

result := Array new.

(self _firstValidOffset) to: values size do:
  [:eachIndex |
    result add: (aBlock value: (values at: eachIndex)).
].

^RcPipe new addAll: result ; yourself
%

category: 'Initializing'
method: RcPipe
initialize: anInitialSize

"initialize the sub-components"

values := Array new: anInitialSize.
consumerState := Association newWithKey: true value: 0.
producerState := Association newWithKey: true value: 0.
%

category: 'Private'
method: RcPipe
_consumer

"Returns whether the receiver performed a removal during the current
 transaction."

^ (System redoLog redoObjects at: consumerState otherwise: nil) ~~ nil
%

category: 'Private'
method: RcPipe
_firstValid

"Private.  Determines if this session is a producer / consumer.  Computes the
 re-synchronized 0-based offset of the first valid element."

| sz |

(self _producer) & (self _consumer) ifTrue: [
  self _trivialCleanUp.
  ^0
].
self _producer ifTrue: [ ^ consumerState value - producerState value ].
self _consumer ifTrue: [ ^consumerState value ].

"neither a producer nor a consumer."
consumerState key = producerState key 
  ifTrue: [ "producer and consumer are in sync"
    sz := consumerState value
  ]
  ifFalse: [ "producer has done a compaction"
    sz := consumerState value - producerState value
  ].
^sz
%

category: 'Private'
method: RcPipe
_firstValidOffset

"Computes the offset of the first valid element in the pipe."

^self _firstValid + 1
%

category: 'Private'
method: RcPipe
_logConsumer

"Logs the fact that a removal occurred."

(System redoLog redoObjects at: consumerState otherwise: nil) == nil
  ifTrue: [ System redoLog redoObjects at: consumerState put: consumerState ]
%

category: 'Private'
method: RcPipe
_logProducer

"Logs the fact that an add occurred."

(System redoLog redoObjects at: producerState otherwise: nil) == nil
  ifTrue: [ System redoLog redoObjects at: producerState put: producerState ]
%

category: 'Private'
method: RcPipe
_performedRefresh

"Returns true if the receiver performed an add / remove during the current
 transaction; returns false otherwise."

^self _producer | self _consumer
%

category: 'Private'
method: RcPipe
_producer

"Returns true if the receiver performed an add during the current
 transaction; returns false otherwise."

^ (System redoLog redoObjects at: producerState otherwise: nil) ~~ nil
%

category: 'Private'
method: RcPipe
_refreshForProducer: aBoolean

"This method is called to re-synchronize the view of a producer or a consumer."

self _performedRefresh ifTrue: [ ^self ]. "do this once per transaction"

System _addRootObjectToRcReadSet: self.
System _addEntireObjectToRcReadSet: values.
System _addRootObjectToRcReadSet: producerState.
System _addRootObjectToRcReadSet: consumerState.

aBoolean == nil ifTrue: [ ^ self ].

aBoolean ifTrue: [ "producer"
  | numToRemove |
  consumerState key = producerState key ifFalse: [ 
    "consumer is out of sync with us. We have to wait
     until he is in sync before attempting further
     compaction"
    ^self
  ].
  "consumer and producer are in sync ; if the consumer has removed any further 
   items ; cleanup; set a flag to indicate cleanup has been performed "
  producerState value: 0.
  numToRemove := consumerState value .
  numToRemove > 0 ifTrue: [
    values removeFrom: 1 to: numToRemove.
    producerState value: numToRemove.
    "indicate to the consumer that a re-sync is necessary"
    producerState key: producerState key not .
  ].
]
ifFalse: [ "consumer"
  | myOffset otherOffset |
  myOffset := consumerState value.
  otherOffset := producerState value.
  consumerState key = producerState key
    ifFalse: [ 
      " the producer is out of sync with us. this means he has just
        done a cleanup. Resynchronize and indicate that resynchronization has
        taken place by setting a flag."
      consumerState value: myOffset - otherOffset.
      consumerState key: consumerState key not.
    ].
].
^self
%

category: 'Private'
method: RcPipe
_setupAdd

"Performs the setup for an add operation (used to avoid read-write conflicts)."

self _refreshForProducer: true.
self _logProducer.
%

category: 'Private'
method: RcPipe
_setupRemove

"Performs the setup for a remove operation"

self _refreshForProducer: false.
self _logConsumer.
%

category: 'Private'
method: RcPipe
_trivialCleanUp

"Private.  Used to cleanup in situations where a single session is
 adding / removing from the queue."

| items |

items := consumerState value - producerState value.
items > values size ifTrue: [ items := values size ].
items > 0 ifTrue: [ 
  values removeFrom: 1 to: items.
  consumerState value: 0.
  producerState value: 0.
  producerState key: consumerState key.
].
%

category: 'Removing'
method: RcPipe
remove

"Removes the leading element from the receiver and returns that element.
 If the receiver is empty, returns nil."

| array | 

array := self removeCount: 1.
array size > 0 ifTrue: [ ^ array at: 1 ].
^nil
%

category: 'Removing'
method: RcPipe
peek

"Returns the leading element from the receiver without removing it.
 If the receiver is empty, returns nil."

| start |

self isEmpty ifTrue: [ ^nil ].
self _setupRemove.
start := self _firstValidOffset.
start > 0 ifTrue: [ ^values at: start ].
^nil
%

category: 'Removing'
method: RcPipe
removeAll

"Removes all entries from the RcPipe, and returns an Array that contains
 those entries, in order."

^self removeCount: self size
%

category: 'Removing'
method: RcPipe
removeCount: maxToRemove

"Removes entries from the RcPipe, and returns an Array that contains the
 minimum of maxToRemove or the queue size entries, in order."

|start end|

maxToRemove == 0 | self isEmpty ifTrue: [ ^#[] ].
self _setupRemove.
start := self _firstValidOffset.
maxToRemove > self size
  ifTrue: [ ^self removeCount: self size ].
end := start + maxToRemove - 1.
consumerState value: (consumerState value + maxToRemove).
^values copyFrom: start to: end.
%

category: 'Testing'
method: RcPipe
isEmpty

"Returns true if the queue is empty, and false otherwise."

^self size == 0
%

category: 'Updating'
method: RcPipe
changeToSegment: segment

"Assigns the receiver and subcomponents to the given segment."

self assignToSegment: segment.
values assignToSegment: segment.
consumerState assignToSegment: segment.
producerState assignToSegment: segment.
%

